#ifndef ASPEN_DFCOMPLEX_H_INCLUDED

#include "aspen_dfreal.h"

struct __align__(16) __cudfcomplex__
{
  cudfreal x, y;
};

typedef	struct __cudfcomplex__	cudfcomplex;


#ifdef __cplusplus

#ifdef __CUDA_ARCH__
#if GPU_ARCH>300
__device__ __forceinline__ cudfcomplex
__ldg ( const cudfcomplex * a )
{
  const cudfreal tx = __ldg( &(a->x) );
  const cudfreal ty = __ldg( &(a->y) );
  const cudfcomplex t = { tx, ty };
  return t;
}
#endif
#endif


__host__ __device__ __forceinline__ bool
operator== ( const cudfcomplex a, const cudfcomplex b )
{
  const bool ret = ((a.x == b.x) && (a.y == b.y)) ? true : false;
  return ret;
}

__host__ __device__ __forceinline__ bool
operator!= ( const cudfcomplex a, const cudfcomplex b )
{
  const bool ret = ((a.x != b.x) || (a.y != b.y)) ? true : false;
  return ret;
}


__host__ __device__ __forceinline__ cudfcomplex
operator+ ( const cudfcomplex a )
{
  return (a);
}

__host__ __device__ __forceinline__ cudfcomplex
operator+ ( const cudfcomplex a, const cudfcomplex b )
{
  const cudfreal tx = a.x + b.x;
  const cudfreal ty = a.y + b.y;
  const cudfcomplex t = { tx, ty };
  return t;
}

__host__ __device__ __forceinline__ void
operator+= ( cudfcomplex &a, const cudfcomplex b )
{
  a = (a + b);
}

__host__ __device__ __forceinline__ cudfcomplex
operator- ( const cudfcomplex a )
{
  const cudfreal tx = -a.x;
  const cudfreal ty = -a.y;
  const cudfcomplex t = { tx, ty };
  return t;
}

__host__ __device__ __forceinline__ cudfcomplex
operator- ( const cudfcomplex a, const cudfcomplex b )
{
  const cudfreal tx = a.x - b.x;
  const cudfreal ty = a.y - b.y;
  const cudfcomplex t = { tx, ty };
  return t;
}

__host__ __device__ __forceinline__ void
operator-= ( cudfcomplex &a, const cudfcomplex b )
{
  a = (a - b);
}

__host__ __device__ __forceinline__ cudfcomplex
operator* ( const cudfcomplex a, const cudfcomplex b )
{
  const cudfreal e = - a.y;
  const cudfreal tx = a.x * b.x;
  const cudfreal ty = a.x * b.y;
  const cudfreal sx = fma( e,   b.y, tx );
  const cudfreal sy = fma( a.y, b.x, ty );
  const cudfcomplex t = { sx, sy };
  return t;
}

__host__ __device__ __forceinline__ cudfcomplex
operator* ( const cudfcomplex a, const cudfreal b )
{
  cudfcomplex t;
  t.x = a.x * b; t.y = a.y * b;
  return t;
}

__host__ __device__ __forceinline__ cudfcomplex
operator* ( const cudfreal a, const cudfcomplex b )
{
  cudfcomplex t = (b * a);
  return t;
}

__host__ __device__ __forceinline__ void
operator*= ( cudfcomplex &a, const cudfcomplex b )
{
  a = (a * b);
}

__host__ __device__ __forceinline__ void
operator*= ( cudfcomplex &a, const cudfreal b )
{
  a = (a * b);
}

__host__ __device__ __forceinline__ cudfcomplex
fma ( const cudfcomplex a, const cudfcomplex b, const cudfcomplex c )
{
  const cudfreal e = - a.y;
  const cudfreal tx = fma( a.x, b.x, c.x );
  const cudfreal ty = fma( a.x, b.y, c.y );
  const cudfreal sx = fma( e,   b.y, tx );
  const cudfreal sy = fma( a.y, b.x, ty );
  const cudfcomplex t = { sx, sy };
  return t;
}

__host__ __device__ __forceinline__ cudfcomplex
fma ( const cudfreal a, const cudfcomplex b, const cudfcomplex c )
{
  const cudfreal tx = fma( a, b.x, c.x );
  const cudfreal ty = fma( a, b.y, c.y );
  const cudfcomplex t = { tx, ty };
  return t;
}

__host__ __device__ __forceinline__ cudfcomplex
fma ( const cudfcomplex a, const cudfreal b, const cudfcomplex c )
{
  cudfcomplex t = fma( b, a, c );
  return t;
}

__host__ __device__ __forceinline__ cudfcomplex
CUWDIV ( const cudfcomplex a, const cudfcomplex b )
{
  /*
   * almost similar approach to divide a complex number as cuComplex.h
   */
  const cudfreal ZERO_ = { (float)0, (float)0 };
  const cudfcomplex ZERO = { ZERO_, ZERO_ };
  if ( b == ZERO ) {
    const cudfreal NAN_DF = { NAN, NAN };
    const cudfcomplex NAN_DFC = { NAN_DF, NAN_DF };
    return NAN_DFC;
  }
  if ( a == ZERO ) { return ZERO; }
  const cudfreal ONE_ = { (float)1, (float)0 };
  const cudfcomplex ONE  = { ONE_, ZERO_ };
  if ( b == ONE  ) { return a; }
  const cudfreal MONE_ = { (float)1, (float)0 };
  const cudfcomplex MONE = { MONE_, ZERO_ };
  if ( b == MONE ) { return -a; }

  cudfreal s = Abs(b.x) + Abs(b.y);
  cudfreal r = ONE_ / s;
  const cudfreal ar = a.x * r;
  const cudfreal ai = a.y * r;
  const cudfreal br = b.x * r;
  const cudfreal bi = b.y * r;
  s = (br * br) + (bi * bi);
  r = ONE_ / s;
  const cudfcomplex t = {
    ((ar * br) + (ai * bi)) * r,
    ((ai * br) - (ar * bi)) * r };
  return t;
}

__host__ __device__ __forceinline__ cudfcomplex
operator/ ( const cudfcomplex a, const cudfcomplex b )
{
  const cudfcomplex t = CUWDIV ( a, b );
  return t;
}

__host__ __device__ __forceinline__ void
operator/= ( cudfcomplex &a, const cudfcomplex b )
{
  a = ( a / b );
}


__host__ __device__ __forceinline__ cudfreal
Abs ( const cudfcomplex a )
{
  const cudfreal r2 = a.x*a.x + a.y*a.y;
  const cudfreal t = Sqrt( r2 );
  return t;
}

__host__ __device__ __forceinline__ cudfcomplex
Conj ( const cudfcomplex a )
{
  const cudfreal tx =  a.x;
  const cudfreal ty = -a.y;
  const cudfcomplex t = { tx, ty };
  return t;
}

__host__ __device__ __forceinline__ cudfcomplex
__choose__ ( const bool flag, const cudfcomplex a, const cudfcomplex b )
{
  const cudfreal tx = __choose__( flag, a.x, b.x );
  const cudfreal ty = __choose__( flag, a.y, b.y );
  const cudfcomplex t = { tx, ty };
  return t;
}

#endif

#define ASPEN_DFCOMPLEX_H_INCLUDED		1
#endif

